#ifndef NVIDIA_H_
#define NVIDIA_H_
#include <class/cl0040.h>
#include <class/cl0080.h>
#include <class/cl2080.h>
#include <nvos.h>
#include <nv_escape.h>
#include <nv-ioctl.h>
#include <nvidia-drm-ioctl.h>
#include <ctrl/ctrl0000/ctrl0000gpu.h>
#include <ctrl/ctrl0000/ctrl0000unix.h>
#include <ctrl/ctrl0000/ctrl0000system.h>
#define NVKMS_API_H
#include <nvkms-api-types.h>
#include <nvkms-kapi-private.h>
#endif

